# EntityCopy

## 1. 개요

EntityCopy는 [DCL](http://www.gowoonsoft.com/dcl)로 개발된 데이터베이스 데이터 복사 유틸리티입니다.

DCL은 Linux를 비롯한 UNIX 운영체제와 Microsoft Windows 운영체제에서 소스프로그램 수준의 이식성있는 C++ 클래스 라이브러리 입니다.

DCL에는 [Firebird](https://www.firebirdsql.org/), [InterBase](https://www.embarcadero.com/products/interbase), [Informix](https://www.ibm.com/kr-ko/products/informix), [MariaDB](https://mariadb.com/docs/server/connect/programming-languages/c/), [MySQL](https://dev.mysql.com/downloads/c-api/), [Oracle Database](https://www.oracle.com/kr/database/technologies/appdev/oci.html) 네이티브 C 프로그래밍 언어 API를 사용한 드라이버 구현을 포함합니다.

EntityCopy는 데이터베이스 스키마에 포함된 테이블 전체 또는 일부를 서로다른 벤더(vendor)의 데이터베이스로 복사할 수 있습니다.

EntityCopy는 데이터베이스간 DDL을 생성하지 않으며, TRUNCATE(or DELETE), SELECT, INSERT를 사용하여 데이터를 단순 복사합니다.

## 2. 다운로드

- Linux x86_64 바이너리
  - entitycopy-1.0-linux-x86_64.tar.gz
- Windows x86_64 바이너리
  - entitycopy-1.0-windows-x86_64.zip
  - entitycopy-1.0-windows-x86_64-bundle.zip
- Source Code
  - entitycopy-1.0-src.zip

## 3. Linux 구성

### 3.1 dcl

배포 패키지의 압축을 풀면 dcl 디렉토리에 다음의 파일이 포함되어 있습니다.
파일명이 'D' 또는 'd'로 종결되는 파일은 DEBUG 빌드된 것을 표시합니다.

bin은 PATH에 포함시키고, lib는 LD_LIBRARY_PATH에 포함시킵니다.

- bin
  - entitycopy, entitycopyD
- lib
  - libDCLCore.so, libDCLCored.so
  - DCLFirebird.so, DCLFirebirdd.so
  - DCLInformix.so, DCLInformixd.so
  - DCLInterBase.so, DCLInterBased.so
  - DCLMariaDB.so, DCLMariaDBd.so
  - DCLOracleDb.so, DCLOracleDbd.so
  - DCLPostgreSQL.so, DCLPostgreSQLd.so

다음은 Linux 환경변수 구성의 예입니다.

~~~sh
export FIREBIRD='/opt/firebird-5.0.2'
export INFORMIXDIR='/opt/informix-csdk-4.50'
export INTERBASE='/opt/interbase-14.6'
export ORACLE_HOME='/opt/oracle-client-21.3'
export PATH=${PATH}:${INFORMIXDIR}/bin:${ORACLE_HOME}/bin
export LD_LIBRARY_PATH=/opt/dcl/lib:${FIREBIRD}/lib:${INFORMIXDIR}/lib:${INFORMIXDIR}/lib/esql:${INTERBASE}/lib:${ORACLE_HOME}/lib
~~~

### 3.2 Firebird

[FIrebird 서버패키지](https://www.firebirdsql.org/en/server-packages/)에서 **libfbclient.so**가 포함된 lib 디렉토리를 LD_LIBRARY_PATH에 포함시킵니다.

### 3.3 Informix

Informix 드라이버는 Informix ESQL/C로 구현되어 있습니다.

[Informix CSDK](https://www.ibm.com/kr-ko/products/informix/developer-tools)를 설치한 후,
\${INFORMIXDIR}/lib/esql 디렉토리를 LD_LIBRARY_PATH에 포함시킵니다.

### 3.4 InterBase

[InterBase 서버패키지](https://www.embarcadero.com/products/interbase)에서 **libgds.so**가 포함된 lib 디렉토리를 LD_LIBRARY_PATH에 포함시킵니다.

### 3.5 MariaDB

MariaDB 드라이버는 [MariaDB Connector/C](https://mariadb.org/connector-c)로 빌드 되었습니다.  
다음과 같이 이를 설치합니다.

~~~sh
# Ubuntu, Debian
apt update; apt install -y libmariadb

# RHEL, Rocky, Oracle Linux
dnf install -y mariadb-connector-c
~~~

### 3.6 Oracle Database

[Oracle Instant Client](https://www.oracle.com/database/technologies/instant-client/linux-x86-64-downloads.html)를 설치 한 후,
**libclntsh.so**가 포함된 디렉토리를 LD_LIBRARY_PATH에 포함시킵니다.

Ubuntu 24.04 noble의 경우 libaio1t64 배포 패키지에 오류가 있습니다.

~~~sh
# dpkg -L libaio1t64
sudo ln -s /usr/lib/x86_64-linux-gnu/libaio.so.1t64.0.2 /usr/lib/x86_64-linux-gnu/libaio.so.1
~~~

### 3.7 PostgreSQL

PostgreSQL 드라이버는 [PostgreSQL Client Interfaces](https://www.postgresql.org/docs/current/client-interfaces.html)로 빌드 되었습니다.  
다음과 같이 이를 설치합니다.

~~~sh
# Ubuntu, Debian
apt update; apt install -y libecpg6

# RHEL, Rocky, Oracle Linux
dnf install -y libecpg
~~~

## 4. Windows 구성

### 4.1 dcl

배포 패키지의 압축을 풀면 dcl 디렉토리에 다음의 파일이 포함되어 있습니다.
파일명이 'D' 또는 'd'로 종결되는 파일은 DEBUG 빌드된 것을 표시합니다.

DCL을 PATH에 포함시킵니다.

- DCL
  - DCLCore.dll, DCLCored.dll
  - DCLFirebird.dll, DCLFirebirdd.dll
  - DCLInformix.dll, DCLInformixd.dll
  - DCLInterBase.dll, DCLInterBased.dll
  - DCLMariaDB.dll, DCLMariaDBd.dll
  - DCLOracleDb.dll, DCLOracleDbd.dll
  - DCLPostgreSQL.dll, DCLPostgreSQLd.dll
  - EntityCopy.exe, EntityCopyD.exe
  - expat.dll, expatd.dll

~~~sh
set DCL=D:\Opt\DCL
set FIREBIRD=D:\Opt\Firebird
set INFORMIXDIR=D:\Opt\informix
set INTERBASE=D:\Opt\InterBase
set MARIADB=D:\Opt\mariadb-connector-c-3.4.3
set ORACLE_HOME=D:\Opt\oracle21_client_home
set POSTGRESQL=D:\Opt\PostgreSQL17
set Path=%Path%;%DCL%;%FIREBIRD%;%INFORMIXDIR%\bin;%INTERBASE%\bin;%MARIADB%\lib64;%ORACLE_HOME%\bin;%POSTGRESQL%\bin
~~~

#### Microsoft Visual C++ Redistributable

배포 패키지는 Visual Studio 2022로 빌드되었습니다.  
필요한 경우 다음에서 다운로드하여 설치합니다.

<https://learn.microsoft.com/en-us/cpp/windows/latest-supported-vc-redist>

### 4.2 Firebird

[FIrebird 서버패키지](https://www.firebirdsql.org/en/server-packages/)에서 **fbclient.dll**이 포함된 디렉토리를 PATH에 포함시킵니다.

### 4.3 Informix

Informix 드라이버는 Informix ESQL/C로 구현되어 있습니다.

[Informix CSDK](https://www.ibm.com/kr-ko/products/informix/developer-tools)를 설치한 후,
\${INFORMIXDIR}\bin 디렉토리를 PATH에 포함시킵니다.

### 4.4 InterBase

[InterBase 서버패키지](https://www.embarcadero.com/products/interbase)에서 **ibclient64.dll**이 포함된 bin 디렉토리를 PATH에 포함시킵니다.

### 4.5 MariaDB

[MariaDB Connector/C](https://mariadb.com/downloads/connectors/) Windows 바이너리를 설치한 후, **libmariadb.dll**이 포함된 lib64 디렉토리를 PATH에 포함시킵니다.

### 4.6 Oracle Database

[Oracle Instant Client](https://www.oracle.com/kr/database/technologies/instant-client/winx64-64-downloads.html)를 설치 한 후, **oci.dll**이 포함된 디렉토리를 PATH에 포함시킵니다.

**legacy.dll, oci.dll, oraociicus.dll** 세 파일만 있어도 동작합니다.

### 4.7 PostgreSQL

PostgreSQL Windows 클라이언트를 별도의 패키지로 배포하지 않는 것 같습니다.

[PostgreSQL Windows installers](https://www.postgresql.org/download/windows/)를 다운로드 한 후, bin 디렉토리를 PATH에 포함시킵니다.

PostgreSQL 드라이버는 libpq, libecpg를 사용하도록 빌드되었습니다. Windows에서 이 외에 필요한 파일은 본 문서의 하단에 별도로로 표기되어 있습니다.

## 5 Execute

~~~sh
EntityCopyD --source="DRIVER=DCLFirebird;USER=TIMETABLE;PASSWORD=TIMETABLE;SERVER=g02u24kn2/32055;DATABASE=TIMETABLE" --destination="DRIVER=DCLInterBase;USER=SYSDBA;PASSWORD=masterkey;SERVER=n0317ud70p/3050;DATABASE=TIMETABLE-UTF8" --truncate --verbose -n

EntityCopyD --source="DRIVER=DCLFirebird;USER=NETIS;PASSWORD=NETIS;SERVER=g02u24kn2/32055;DATABASE=NETIS" --destination="DRIVER=DCLInterBase;USER=SYSDBA;PASSWORD=masterkey;SERVER=n0317ud70p/3050;DATABASE=NETIS-UTF8" --truncate --rows=100000 --verbose -n

EntityCopyD --source="DRIVER=DCLOracleDb;USER=MT100;PASSWORD=MT100;DATABASE=g02r08od21:1521/AL32UTF8" --destination="DRIVER=DCLInterBase;USER=SYSDBA;PASSWORD=masterkey;SERVER=n0317ud70p/3050;DATABASE=MT100" --truncate --rows=100000  --i8-precision=20 --verbose -n

entitycopyD \
    --source="DRIVER=DCLFirebird;USER=TIMETABLE;PASSWORD=TIMETABLE;SERVER=g02u24kn2/32055;DATABASE=TIMETABLE" \
    --destination="DRIVER=DCLInterBase;USER=SYSDBA;PASSWORD=masterkey;SERVER=n0317ud70p/3050;DATABASE=TIMETABLE-UTF8" \
    --truncate --verbose -n

entitycopyD \
    --source="DRIVER=DCLFirebird;USER=NETIS;PASSWORD=NETIS;SERVER=g02u24kn2/32055;DATABASE=NETIS" \
    --destination="DRIVER=DCLInterBase;USER=SYSDBA;PASSWORD=masterkey;SERVER=n0317ud70p/3050;DATABASE=NETIS-UTF8" \
    --truncate --rows=10000 --verbose -n
~~~

## Distriute

~~~sh
# Windows
del DCL\*

copy ..\bin\EntityCopy*.exe DCL\
copy ..\lib\expat*.dll DCL\
copy ..\lib\DCL*.dll DCL\

7z a entitycopy-1.0-windows-x86_64.zip DCL

copy %FIREBIRD%\fbclient.dll DCL\

copy %INFORMIXDIR%\bin\esqlauth.dll DCL\
copy %INFORMIXDIR%\bin\igl4n304.dll DCL\
copy %INFORMIXDIR%\bin\iglxn304.dll DCL\
copy %INFORMIXDIR%\bin\igo4n304.dll DCL\
copy %INFORMIXDIR%\bin\iregt07b.dll DCL\
copy %INFORMIXDIR%\bin\irrgt09a.dll DCL\
copy %INFORMIXDIR%\bin\irset09a.dll DCL\
copy %INFORMIXDIR%\bin\isqlt09a.dll DCL\

copy %INTERBASE%\bin\ibclient64.dll DCL\

copy %MARIADB%\lib64\libmariadb.dll DCL\

set ORACLE_CLIENT=D:\Opt\downloads\instantclient_23_8
copy %ORACLE_CLIENT%\legacy.dll DCL\
copy %ORACLE_CLIENT%\oci.dll DCL\
copy %ORACLE_CLIENT%\oraociicus.dll DCL\

copy %POSTGRESQL%\bin\libcrypto-3-x64.dll DCL\
copy %POSTGRESQL%\bin\libecpg.dll DCL\
copy %POSTGRESQL%\bin\libiconv-2.dll DCL\
copy %POSTGRESQL%\bin\libintl-9.dll DCL\
copy %POSTGRESQL%\bin\libpgtypes.dll DCL\
copy %POSTGRESQL%\bin\libpq.dll DCL\
copy %POSTGRESQL%\bin\libssl-3-x64.dll DCL\
copy %POSTGRESQL%\bin\libwinpthread-1.dll DCL\

7z a entitycopy-1.0-windows-x86_64-bundle.zip DCL

# Linux
rm -rf dcl; mkdir -p dcl/bin dcl/lib
cp -a ../bin/entitycopy* dcl/bin/
cp -a ../lib/libDCL* dcl/lib/
cp -a ../lib/DCL* dcl/lib/

tar --owner=0 --group=0 --numeric-owner -czf dcl-entitycopy-1.0-linux-x86_64.tar.gz *
~~~
